import heapq
from typing import List, Optional

import matplotlib.pyplot as plt
import networkx as nx
from networkx import NetworkXNoPath
from shapely import LineString, MultiLineString, Point
from shapely.ops import linemerge

from imxInsights.graph.imxGraphModels import (
    DirectionEnum,
    GraphRoute,
    MicroLinkConnection,
    MicroLinkConnectionRepo,
    MicroLinkRepo,
    MicroNodeRepo,
    PathEdge,
    PathEdgeData,
    RailConnectionRepo,
)
from imxInsights.graph.imxGraphSituationProtocol import ImxSituationRepoMinimalProtocol
from imxInsights.utils.log import logger
from imxInsights.utils.shapely_helpers import cut, cut_profile, offset_linestring


class ImxGraph:
    def __init__(
        self,
        graph: nx.DiGraph,
        imx_situation: ImxSituationRepoMinimalProtocol,
        micro_link_repo: MicroLinkRepo,
        micro_node_repo: MicroNodeRepo,
        rail_connections_repo: RailConnectionRepo,
        micro_link_connections: MicroLinkConnectionRepo,
    ):
        self.g = graph
        self.imx_situation = imx_situation
        self.micro_link_repo = micro_link_repo
        self.micro_node_repo = micro_node_repo
        self.rail_connections_repo = rail_connections_repo
        self.micro_link_connections = micro_link_connections
        self._path_finding_cash = {}

    def _get_plot(self):
        fig, ax = plt.subplots()
        length = 2
        for edge in self.g.edges:
            edge_data = self.g.get_edge_data(edge[0], edge[1])
            line_source = self.micro_link_repo.get_by_puic(edge[0]).get_shapely()
            line_target = self.micro_link_repo.get_by_puic(edge[1]).get_shapely()

            if edge_data["current_direction"] == "UPSTREAM":
                point_1 = line_source.interpolate(length)
            elif edge_data["current_direction"] == "DOWNSTREAM":
                point_1 = line_source.interpolate(line_source.length - length)
            else:
                raise ValueError()

            if edge_data["next_direction"] == "UPSTREAM":
                point_2 = line_target.interpolate(line_target.length - length)
            elif edge_data["next_direction"] == "DOWNSTREAM":
                point_2 = line_target.interpolate(length)
            else:
                raise ValueError()

            edge_line = offset_linestring(LineString([Point(point_1.x, point_1.y), Point(point_2.x, point_2.y)]), 0.25, "down")
            plt.annotate("", xy=edge_line.coords[0], xytext=edge_line.coords[-1], arrowprops=dict(facecolor="red", arrowstyle="->"))

        for node in self.g.nodes:
            line = self.micro_link_repo.get_by_puic(node).get_shapely()
            x, y = line.xy
            ax.plot(x, y, linewidth=7.0)

        plt.axis("equal")
        return fig

    def draw_graph(self, path: Optional[List[str]] = None):
        """Draws the graph, optionally highlighting a specified path.

        Args:
            path (Optional[List[str]]): A list of PUICs representing the path to highlight. If None, no path is highlighted.
        """
        plt = self._get_plot()
        plt.show()

    def get_edges_path(self, path: List[str], data: bool = True) -> List[PathEdge]:
        edges = []
        for i in range(len(path) - 1):
            source = path[i]
            target = path[i + 1]
            if data:
                edge_data = self.g.get_edge_data(source, target)
                # Assuming edge_data contains keys that exactly match the PathEdgeData fields
                # and values that can be directly used to instantiate PathEdgeData
                path_edge_data = PathEdgeData(
                    source_direction=DirectionEnum[edge_data["current_direction"]],
                    target_direction=DirectionEnum[edge_data["next_direction"]],
                    junction_ref=edge_data["junction"],
                    graph_junction=edge_data["graph_junction"],
                    jumper=edge_data["jumper"],
                )
                edge = PathEdge(source, target, path_edge_data)
            else:
                edge = PathEdge(source, target)
            edges.append(edge)
        return edges

    @staticmethod
    def has_cutback_in_path(edges_in_path: List[PathEdge]) -> bool:
        cutback_in_route = False
        for idx, item in enumerate(edges_in_path):
            if idx < len(edges_in_path) - 1:
                if item.data.junction_ref == edges_in_path[idx + 1].data.junction_ref:
                    cutback_in_route = True
        return cutback_in_route

    @staticmethod
    def check_if_not_multiline(rail_con_geometry: List[LineString]) -> bool:
        tester = linemerge(rail_con_geometry)
        if isinstance(tester, MultiLineString):
            return False
        return True

    def get_objects_on_route(
        self,
        micro_link_ref: str,
        measure: Optional[float] = None,
    ) -> List[MicroLinkConnection]:
        if measure:
            return [_ for _ in self.micro_link_connections.get_by_puic(micro_link_ref) if _.measure <= measure]
        else:
            return [_ for _ in self.micro_link_connections.get_by_puic(micro_link_ref)]

    @staticmethod
    def append_objects_on_route(direction: DirectionEnum, objects_on_path, object_on_route):
        if direction == DirectionEnum.UPSTREAM:
            objects_on_path.extend(reversed(object_on_route))
        elif direction == DirectionEnum.DOWNSTREAM:
            objects_on_path.extend(object_on_route)

    def cut_geometry(self, path_edge, mirco_link, measure, target=False):
        line_source = self.micro_link_repo.get_by_puic(mirco_link).get_shapely()
        direction = path_edge.data.source_direction if not target else path_edge.data.target_direction
        line_cut = cut(line_source, measure)

        if measure == 0 or line_source.length == line_cut[0].length:
            return line_cut[0]
        elif direction == DirectionEnum.DOWNSTREAM:
            return line_cut[1] if not target else line_cut[0]
        elif direction == DirectionEnum.UPSTREAM:
            return line_cut[0] if not target else line_cut[1]
        else:
            raise ValueError("should have at least one cut on start mirco_link")

    def _find_paths_without_revisiting_nodes(self, G, source, target, weight="weight", max_weight_increase=0.1, cutoff=15):
        """
        Finds paths from source to target in graph G without revisiting nodes and with a length cutoff.

        Only considers paths with a total weight within 10% of the shortest path's weight and under the specified cutoff.

        Parameters:
        - G: The graph
        - source: Source node
        - target: Target node
        - weight: The edge attribute to use as weight
        - max_weight_increase: Maximum allowed weight increase over the shortest path (as a fraction)
        - cutoff: Maximum number of edges (hops) a path can have

        Returns:
        - A list of tuples, where each tuple contains a path and its weight.
        """
        # Cache key based on function parameters that affect the output
        cache_key = (source, target, weight, max_weight_increase, cutoff)

        # Check if the result is in cache
        if cache_key in self._path_finding_cash.keys():
            return self._path_finding_cash[cache_key]

        # Find the shortest path weight to determine the threshold
        try:
            shortest_path_weight = nx.dijkstra_path_length(G, source, target, weight=weight)
        except Exception as e:
            logger.warning(f"no path {e}")
            return []

        weight_threshold = shortest_path_weight * (1 + max_weight_increase)

        # Initialize the priority queue with the starting point
        paths_queue = [(0, [source], {source})]  # Each item: (total_weight, path, visited_nodes)

        # Store valid paths meeting the criteria
        valid_paths = []

        while paths_queue:
            current_weight, current_path, visited = heapq.heappop(paths_queue)
            last_node = current_path[-1]

            # Check if current path meets the target and weight criteria
            if last_node == target and current_weight <= weight_threshold:
                valid_paths.append((current_path, current_weight))

            # If cutoff is defined and the path length exceeds it, skip further extension
            if cutoff is not None and len(current_path) > cutoff:
                continue

            # Explore paths from the current node to its neighbors
            for neighbor in G[last_node]:
                if neighbor not in visited or neighbor == target:
                    new_weight = current_weight + G.edges[last_node, neighbor][weight]

                    # Continue only if the new path weight is within the threshold
                    if new_weight <= weight_threshold:
                        new_path = current_path + [neighbor]
                        new_visited = visited.copy()
                        new_visited.add(neighbor)

                        heapq.heappush(paths_queue, (new_weight, new_path, new_visited))
        # Cache the result before returning
        self._path_finding_cash[cache_key] = valid_paths
        return valid_paths

    def get_path(self, source_mirco_link, target_micro_link, start_measure, end_measure, cutoff=15) -> List[GraphRoute]:
        # todo: make cutoff should be input from query layer

        logger.trace(f"Routes finding between source: {source_mirco_link} {start_measure}, target: {target_micro_link} {end_measure}")
        output = []

        if source_mirco_link == target_micro_link:
            if source_mirco_link in self.micro_link_connections.get_keys():
                objects_on_route = [
                    _ for _ in self.micro_link_connections.get_by_puic(source_mirco_link) if _.measure >= start_measure and _.measure <= end_measure
                ]
                line_profile = cut_profile(self.micro_link_repo.get_by_puic(source_mirco_link).get_shapely(), start_measure, end_measure)
                return [GraphRoute([PathEdge(source_mirco_link, source_mirco_link)], line_profile, objects_on_route)]
            else:
                raise ValueError("MicroLink not in repo")

        # todo implement a own algorithm to find paths

        try:
            all_paths = list(nx.all_shortest_paths(self.g, source=source_mirco_link, target=target_micro_link, weight="weight"))  # , cutoff=cutoff
        except NetworkXNoPath:
            logger.warning(f"Found no routes between {source_mirco_link} and {target_micro_link}.")
            all_paths = []
        logger.info(f"Found {len(all_paths)} routes between {source_mirco_link} and {target_micro_link}.")

        tester = self._find_paths_without_revisiting_nodes(self.g, source_mirco_link, target_micro_link, weight="weight", max_weight_increase=0.1)
        all_paths = [_[0] for _ in tester]
        logger.info(f"(custom)) Found {len(all_paths)} routes between {source_mirco_link} and {target_micro_link}.")

        for path in all_paths:
            edges_in_path = self.get_edges_path(path)

            if self.has_cutback_in_path(edges_in_path):
                pass
                # logger.info(f"Route has cutback, not supported yet, will skip {edges_in_path}")
                # todo: cutback breaks below.... do we want to support this?
                # - objects on route: append them again in reversed order
                # - geometry: reverse cutback part

            else:
                logger.info(f"Processing route {edges_in_path}")

                objects_on_route = []  # this should be on route.... path edge is path.
                rail_con_geometry = []
                for idx, path_edge in enumerate(edges_in_path):
                    if idx == 0:
                        # first part of route, check measure what objects should be included.
                        object_in_route_segment = self.get_objects_on_route(path_edge.source, start_measure)
                        self.append_objects_on_route(path_edge.data.source_direction, objects_on_route, object_in_route_segment)

                        new_line = self.cut_geometry(path_edge, source_mirco_link, start_measure, target=False)
                        rail_con_geometry.append(new_line)

                        # all so can be the last part of route, check measure what objects should be included.
                        if len(edges_in_path) == 1:
                            object_in_route_segment = self.get_objects_on_route(path_edge.target, end_measure)
                            self.append_objects_on_route(path_edge.data.target_direction, objects_on_route, object_in_route_segment)

                            new_line = self.cut_geometry(path_edge, target_micro_link, end_measure, target=True)
                            rail_con_geometry.append(new_line)

                    elif len(edges_in_path) == idx + 1:
                        # last part of route, source all objects are included.
                        if path_edge.source in self.micro_link_connections.get_keys():
                            self.append_objects_on_route(
                                path_edge.data.source_direction, objects_on_route, self.micro_link_connections.get_by_puic(path_edge.source)
                            )
                        rail_con_geometry.append(self.micro_link_repo.get_by_puic(path_edge.source).get_shapely())

                        # target micro_link, check measure what objects should be included.
                        object_in_route_segment = self.get_objects_on_route(path_edge.target, end_measure)
                        self.append_objects_on_route(path_edge.data.target_direction, objects_on_route, object_in_route_segment)

                        # cut geometry
                        new_line = self.cut_geometry(path_edge, target_micro_link, end_measure, target=True)
                        rail_con_geometry.append(new_line)

                    else:
                        if path_edge.source in self.micro_link_connections.get_keys():
                            self.append_objects_on_route(
                                path_edge.data.source_direction, objects_on_route, self.micro_link_connections.get_by_puic(path_edge.source)
                            )
                        rail_con_geometry.append(self.micro_link_repo.get_by_puic(path_edge.source).get_shapely())

                output.append(GraphRoute(edges_in_path, linemerge(rail_con_geometry), objects_on_route))

        return output

    def get_paths_between_imx_objects(self, from_obj, to_obj) -> List[GraphRoute]:
        source_mirco_link = from_obj.rail_connection_infos.rail_infos[0].ref
        target_micro_link = to_obj.rail_connection_infos.rail_infos[0].ref
        start_measure = from_obj.rail_connection_infos.rail_infos[0].at_measure
        end_measure = to_obj.rail_connection_infos.rail_infos[0].at_measure

        paths = self.get_path(source_mirco_link, target_micro_link, start_measure, end_measure)
        return paths



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['imxInsights',
 'imxInsights.diff',
 'imxInsights.domain',
 'imxInsights.domain.area',
 'imxInsights.domain.models',
 'imxInsights.graph',
 'imxInsights.graph.queries',
 'imxInsights.repo',
 'imxInsights.repo.tree',
 'imxInsights.report',
 'imxInsights.utils']

package_data = \
{'': ['*']}

install_requires = \
['loguru',
 'ruamel.yaml',
 'lxml',
 'shapely',
 'shapelyM',
 'pyproj',
 'dateparser',
 'pandas',
 'xlsxwriter',
 'Jinja2',
 'networkx',
 'matplotlib',
 'nest-asyncio',
 'arcGisFeatureCash',
 'nest_asyncio']

extras_require = \
{'dev': ['black[jupyter]',
         'bumpversion',
         'flake8-black',
         'flake8-docstrings',
         'flake8',
         'isort',
         'mypy',
         'pre-commit',
         'pytest',
         'pytest-cov',
         'pytest-mock',
         'markdown-include',
         'mike',
         'mkdocs',
         'mkdocs-jupyter',
         'mkdocs-material',
         'mkdocs-mermaid2-plugin',
         'mkdocstrings[crystal, python]>=0.18',
         'pymdown-extensions',
         'python-dotenv',
         'cryptography',
         'matplotlib'],
 'jupyter': ['jupyterlab', 'pygwalker']}

setup(name='imxInsights',
      version='0.1.0.dev1',
      description='python imx insights module to get information from imx files',
      author=None,
      author_email='Hzd <Hazedd@users.noreply.github.com>',
      url=None,
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='~=3.10',
     )



from __future__ import annotations

import math
from typing import List, Union

import numpy as np
import pyproj
from shapely.geometry import LineString, Point, Polygon


def gml_point_to_shapely(gml_point_coordinates: str) -> Point:
    """
    Converts a GML point coordinate string to a Shapely Point object.

    Args:
        gml_point_coordinates (str): The GML point coordinate string in the format "(x,y)".

    Returns:
        (Shapely.Point): The Shapely Point object.

    """
    coordinates = [float(x) for x in gml_point_coordinates.replace("(", "").replace(")", "").replace("'", "").replace(" ", "").split(",")]
    return Point(coordinates)


def shapely_point_to_gml(shapely_point: Point):
    """
    Converts a Shapely Point object to a GML point coordinate string.

    Args:
        shapely_point (Point): The Shapely point".

    Returns:
        (str): gml:point string representation.

    """
    if shapely_point.has_z:
        return f"{round(shapely_point.x, 3)},{round(shapely_point.y, 3)},{round(shapely_point.z, 3)}"
    else:
        return f"{round(shapely_point.x, 3)},{round(shapely_point.y, 3)}"


def gml_linestring_to_shapely(gml_linestring_coordinates: str) -> LineString:
    """
    Converts a GML linestring coordinate string to a Shapely LineString object.

    Args:
        gml_linestring_coordinates (str): A string of GML linestring coordinates in "x,y" format separated by spaces.

    Returns:
        (Shapely.LineString): A Shapely LineString object.

    """
    return LineString([tuple(map(float, x.split(","))) for x in gml_linestring_coordinates.split(" ")])


def gml_polygon_to_shapely(gml_linestring_coordinates: str) -> Polygon:
    """
    Converts a GML polygon to a Shapely Polygon object.

    Args:
        gml_linestring_coordinates (str): A string containing the GML coordinates of the polygon.

    Returns:
        (Polygon): A Shapely Polygon object.

    """
    return Polygon([tuple(map(float, x.split(","))) for x in gml_linestring_coordinates.split(" ")])


class ShapelyTransform:
    """A utility class to transform between RD and WGS84 coordinate systems."""

    rd = pyproj.CRS("EPSG:28992")
    wgs = pyproj.CRS("EPSG:4326")
    transformer_to_wgs = pyproj.Transformer.from_crs(rd, wgs)
    transformer_to_rd = pyproj.Transformer.from_crs(wgs, rd)

    @classmethod
    def rd_to_wgs(cls, shapely: Union[Point, LineString, Polygon]) -> Union[Point, LineString, Polygon]:
        """
        Convert a Shapely geometry from Dutch RD (Rijksdriehoekstelsel) coordinates (EPSG:28992) to WGS84 coordinates (EPSG:4326).

        Args:
            shapely (Union[Point, LineString, Polygon]): A Shapely geometry in Dutch RD coordinates.

        Returns:
            (Union[Point, LineString, Polygon]): A Shapely geometry in WGS84 coordinates.

        """
        return cls._convert(shapely, cls.transformer_to_wgs)

    @staticmethod
    def _convert(shapely: Union[Point, LineString, Polygon], transformer: pyproj.Transformer) -> Union[Point, LineString, Polygon]:
        if isinstance(shapely, Point):
            return Point(*reversed(transformer.transform(shapely.x, shapely.y)))

        elif isinstance(shapely, LineString):
            return LineString(zip(*reversed(transformer.transform(*shapely.coords.xy))))

        elif isinstance(shapely, Polygon):
            return LineString(zip(*reversed(transformer.transform(*shapely.exterior.coords.xy))))
        else:
            return shapely


def reverse_line(shapely_polyline: LineString) -> LineString:
    """
    Reverses the order of coordinates in a Shapely LineString object.

    Args:
        shapely_polyline (LineString): The LineString object to reverse.

    Returns:
        (LineString): A new LineString object with the coordinates in reverse order.

    """
    return LineString(list(shapely_polyline.coords)[::-1])


def get_azimuth_from_points(point1: Point, point2: Point) -> float:
    """
    Calculates the azimuth angle between two points.

    Args:
        point1 (Point): The first Point object.
        point2 (Point): The second Point object.

    Returns:
        (float): The azimuth angle in degrees.

    """
    angle = np.arctan2(point2.x - point1.x, point2.y - point1.y)
    return float(np.degrees(angle)) if angle >= 0 else float(np.degrees(angle) + 360)


def check_point_in_area(point_str: str, area: Polygon):
    point_to_test = Point([float(item) for item in point_str.split(",")])
    return point_to_test.within(area)


def cut(line: LineString, distance: float) -> List[LineString]:
    # todo: if 0 then its not working... sho
    if distance == 0:
        return [line]

    # todo: move to shapley helpers
    # Cuts a 3d line in two at a distance from its starting point
    if distance <= 0.0 or distance >= line.length:
        return [LineString(line)]
    coordinates = list(line.coords)
    for i, p in enumerate(coordinates):
        pd = line.project(Point(p))
        if pd == distance:
            return [LineString(coordinates[: i + 1]), LineString(coordinates[i:])]
        if pd > distance:
            cp = line.interpolate(distance)
            # todo: check if z present, if not switch and do not use z
            return [LineString(coordinates[:i] + [(cp.x, cp.y, cp.z)]), LineString([(cp.x, cp.y, cp.z)] + coordinates[i:])]


def cut_profile(line: LineString, measure_from: float, measure_to: float) -> LineString:
    if measure_from > measure_to:
        measure_from, measure_to = measure_to, measure_from
    if measure_from == 0:
        new_line = line
    else:
        new_line = cut(line, measure_from)[1]

    point = line.interpolate(measure_to)
    new_measure = new_line.project(point)
    result = cut(new_line, new_measure)[0]
    return result


def offset_linestring(line: LineString, distance: float, direction: str) -> LineString:
    # Extract coordinates from LineString
    coords = np.array(line.coords)

    # Calculate direction vector
    direction_vector = np.array([coords[-1][0] - coords[0][0], coords[-1][1] - coords[0][1]])
    direction_vector = direction_vector / np.linalg.norm(direction_vector)

    # Calculate perpendicular vector
    if direction == "up":
        perpendicular_vector = np.array([-direction_vector[1], direction_vector[0]])  # Rotate by 90 degrees clockwise
    elif direction == "down":
        perpendicular_vector = np.array([direction_vector[1], -direction_vector[0]])  # Rotate by 90 degrees counterclockwise
    else:
        raise ValueError("Direction must be 'up' or 'down'.")

    offset_coords = []
    for coord in coords:
        new_coord = coord + distance * perpendicular_vector
        offset_coords.append((new_coord[0], new_coord[1]))

    return LineString(offset_coords)


def extend_line(start_point, end_point, extension_meters, extend_from="both"):
    # todo: add typehints
    """
    Extends a line defined by start and end points by a certain number of meters in specified direction(s).

    :param start_point: Tuple (x, y) representing the start point of the line.
    :param end_point: Tuple (x, y) representing the end point of the line.
    :param extension_meters: Number of meters to extend the line.
    :param extend_from: Direction to extend the line ('start', 'end', 'both').
    :return: A Shapely LineString object representing the extended line.
    """
    # Convert points to Shapely Points
    start = Point(start_point)
    end = Point(end_point)

    # Calculate the direction vector and its magnitude
    dx = end.x - start.x
    dy = end.y - start.y
    length = math.sqrt(dx**2 + dy**2)

    # Normalize the direction vector
    dx_normalized = dx / length
    dy_normalized = dy / length

    # Initialize new start and end points as the original ones
    new_start = Point(start.x, start.y)
    new_end = Point(end.x, end.y)

    # Calculate the extension in terms of the line's coordinate system
    extension_dx = dx_normalized * extension_meters
    extension_dy = dy_normalized * extension_meters

    # Extend the line as specified
    if extend_from == "start" or extend_from == "both":
        new_start = Point(start.x - extension_dx, start.y - extension_dy)
    if extend_from == "end" or extend_from == "both":
        new_end = Point(end.x + extension_dx, end.y + extension_dy)

    # Create and return the extended LineString
    extended_line = LineString([new_start, new_end])
    return extended_line


def point_relative_to_linestring(linestring, point):
    # todo: add typehints
    """
    Determines if a point is to the left or right of a LineString considering its drawing direction.

    :param linestring: Shapely LineString object.
    :param point: Shapely Point object to check.
    :return: 'left', 'right', or 'on' indicating the point's position relative to the LineString.
    """
    # Ensure the LineString has at least two points to define a direction
    if len(linestring.coords) < 2:
        raise ValueError("LineString must have at least two points.")

    # Use the first segment of the LineString to determine direction
    start_point = Point(linestring.coords[0])
    direction_point = Point(linestring.coords[1])

    # Vector from start_point to direction_point
    vector_a = (direction_point.x - start_point.x, direction_point.y - start_point.y)

    # Vector from start_point to the input point
    vector_b = (point.x - start_point.x, point.y - start_point.y)

    # Cross product of vector_a and vector_b
    cross_product = vector_a[0] * vector_b[1] - vector_a[1] * vector_b[0]

    # Determine position based on the sign of the cross product
    if cross_product > 0:
        return "left"
    elif cross_product < 0:
        return "right"
    else:
        return "on"


def calculate_vector(linestring):
    """Calculate vector from a linestring."""
    start_point = np.array(linestring.coords[0])
    end_point = np.array(linestring.coords[-1])
    vector = end_point - start_point
    return vector


def normalize_vector(vector):
    """Normalize a vector."""
    norm = np.linalg.norm(vector)
    if norm == 0:
        return vector
    return vector / norm


def are_linestrings_in_same_direction(linestrings):
    """Check if all linestrings are in the same direction."""
    vectors = [calculate_vector(ls) for ls in linestrings]
    normalized_vectors = [normalize_vector(v) for v in vectors]

    # Compare vectors by dot product
    reference_vector = normalized_vectors[0]
    for vector in normalized_vectors[1:]:
        dot_product = np.dot(reference_vector, vector)
        if dot_product < 0:  # Vectors are in opposite directions
            return False
    return True


def reproject_point_to_line(point: Point, line: LineString) -> Point:
    return line.interpolate(line.project(point))


# def shared_segments(line1: LineString, line2: LineString) -> Tuple[LineString, LineString]:
#     segments1 = [LineString((line1.coords[i], line1.coords[i + 1])) for i in range(len(line1.coords) - 1)]
#     segments2 = [LineString((line2.coords[i], line2.coords[i + 1])) for i in range(len(line2.coords) - 1)]
#
#     shared1 = [_ for _ in segments1 if _.intersects(line2)]
#     shared2 = [_ for _ in segments2 if _.intersects(line1)]
#
#     if len(shared1) != 1 or len(shared2) != 1:
#         raise ValueError("None, or more shared")
#
#     return shared1[0], shared2[0]
#
#
# def get_farthest_points(line1: LineString, line2: LineString) -> Tuple[Tuple[float, float], Tuple[float, float]]:
#     # Find the point on line1 farthest from line2
#     farthest_point_line1 = max(line1.coords, key=lambda p: line2.distance(Point(p[0], p[1])))
#
#     # Find the point on line2 farthest from line1
#     farthest_point_line2 = max(line2.coords, key=lambda p: line1.distance(Point(p[0], p[1])))
#
#     return farthest_point_line1, farthest_point_line2



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, List


@dataclass
class ImxReffedObjects:
    """A dataclass representing a collection of ImxRelation objects, object is iterable."""

    objects: List[ImxRelation] = field(default_factory=list)
    has_invalid_relations: bool = False

    def __iter__(self):
        for related_object in self.objects:
            yield related_object

    def __repr__(self):
        return f"<{self.__class__.__name__} {len(self.objects)} has_invalid_relations={self.has_invalid_relations}>"


@dataclass
class ImxRelation:
    """A data class representing a relationship between two objects."""

    type: str
    key: str
    reffed_object: Any
    present_in_dataset: bool = True

    def __post_init__(self):
        if self.reffed_object is None:
            self.present_in_dataset = False



from typing import Dict, List, Optional

import networkx as nx

from imxInsights.graph.imxGraph import ImxGraph
from imxInsights.graph.imxGraphJunction import ImxGraphJunction
from imxInsights.graph.imxGraphModels import (
    DirectionEnum,
    GraphMicroLink,
    ImxMicroLink,
    ImxMicroNode,
    ImxRailConnection,
    MicroLinkConnection,
    MicroLinkConnectionRepo,
    MicroLinkRepo,
    MicroNodeRepo,
    RailConnectionRepo,
)
from imxInsights.graph.imxGraphSituationProtocol import ImxSituationRepoMinimalProtocol


class ImxGraphBuilder:
    """Builds a directed graph (DiGraph) from IMX insight data.

    Attributes:
        imx_situation (SituationRepo): The repository containing situation data.
        g (nx.DiGraph): The directed graph object being built.
        micro_link_repo (MicroLinkRepo): Repository of micro links.
        micro_node_repo (MicroNodeRepo): Repository of micro nodes.
        rail_connections_repo (RailConnectionRepo): Repository of rail connections.
        micro_link_connections (MicroLinkConnectionRepo): Micro link connections extracted from the data.
    """

    def __init__(self, imx_situation: ImxSituationRepoMinimalProtocol):
        """Initializes the GraphBuilder with a situation repository.

        Args:
            imx_situation (SituationRepo): The repository containing situation data.
        """
        self.imx_situation = imx_situation
        self.g = nx.DiGraph()
        self.micro_link_repo = self._get_micro_link_repo()
        self.micro_node_repo = self._get_micro_node_repo()
        self.rail_connections_repo = self._get_rail_connection_repo()
        self.micro_link_connections = self._get_micro_link_connections_repo()

    def _get_micro_link_repo(self) -> MicroLinkRepo:
        """Creates a micro link repository from the situation data.

        Returns:
            MicroLinkRepo: The micro link repository.
        """
        return MicroLinkRepo([ImxMicroLink(_) for _ in self.imx_situation.get_by_types(["MicroLink"])])

    def _get_micro_node_repo(self) -> MicroNodeRepo:
        """Creates a micro node repository from the situation data.

        Returns:
            MicroNodeRepo: The micro node repository.
        """
        return MicroNodeRepo([ImxMicroNode(_) for _ in self.imx_situation.get_by_types(["MicroNode"])])

    def _get_rail_connection_repo(self) -> RailConnectionRepo:
        """Creates a rail connection repository from the situation data.

        Returns:
            RailConnectionRepo: The rail connection repository.
        """
        return RailConnectionRepo([ImxRailConnection(_) for _ in self.imx_situation.get_by_types(["RailConnection"])])

    def _get_linked_micro_links(self, input_micro_link: ImxMicroLink, node_type: str) -> Dict[str, GraphMicroLink]:
        """Finds micro links connected to a given micro link through 'from_node' or 'to_node'.

        This method processes the specified 'from_node' or 'to_node' of an input micro link to find all other micro links that
        are connected through jumpers, which are essentially connections between nodes. It considers the passages refs
        and the transferability of jumpers, their directionality, and the presence of passages to establish connections.

        Args:
            input_micro_link (ImxMicroLink): The micro link to find connections for.
            node_type (str): The node type to consider for connections ('from_node' or 'to_node').

        Returns:
            Dict[str, GraphMicroLink]: A dictionary mapping the input micro link's PUIC to a list of connected GraphMicroLinks.
        """
        output = []  # Initialize an empty list to store the connected GraphMicroLinks.

        # Get the specified 'from_node' or 'to_node' of the input micro link.
        to_or_from_node = getattr(input_micro_link, node_type)

        # Retrieve the micro node object using the reference from the 'to_or_from_node'.
        micro_node = self.micro_node_repo.get_by_puic(to_or_from_node.ref)
        imx_implementation = self.imx_situation.get_by_puic(to_or_from_node.ref)
        graph_junction = ImxGraphJunction.from_element(imx_implementation.element)

        # Iterate through each jumper (connection) of the micro node.
        for jumper in micro_node.jumpers:
            # Skip the jumper if it is not traversable.
            if not jumper.is_traversible:
                continue

            # Filter the passages that are present and not present in the input micro link and in the jumper.
            passages_from_input_micro_link = [_ for _ in jumper.passage_refs if _ in input_micro_link.passage_refs]
            passages_from_other_micro_link = [_ for _ in jumper.passage_refs if _ not in input_micro_link.passage_refs]

            # Check if there are any passages from the input micro link present in the jumper.
            if passages_from_input_micro_link:
                # Iterate through all micro links in the repository to find potential connections.
                for micro_link in self.micro_link_repo.micro_links:
                    for passage_ref in micro_link.passage_refs:
                        # If a passage reference matches and it's not the input micro link, we have a connection.
                        if passage_ref in passages_from_other_micro_link and micro_link != input_micro_link:
                            # Skip if the jumper is not two-way and the connection does not match the current node processing logic.
                            if not jumper.is_two_way:
                                if node_type == "to_node" and micro_link.from_node.ref == micro_node.junctionRef:
                                    continue
                                if node_type == "from_node" and micro_link.to_node.ref == micro_node.junctionRef:
                                    continue

                            # Determine the direction of the connection based on the node references.
                            to_direction = DirectionEnum.DOWNSTREAM if micro_link.from_node.ref == to_or_from_node.ref else DirectionEnum.UPSTREAM

                            # Add the connected GraphMicroLink to the output list.
                            output.append(
                                GraphMicroLink(
                                    imx_micro_link=micro_link,
                                    from_direction=DirectionEnum.UPSTREAM if node_type == "from_node" else DirectionEnum.DOWNSTREAM,
                                    to_direction=to_direction,
                                    junction=micro_node,
                                    graph_junction=graph_junction,
                                    jumper=jumper,
                                )
                            )

        # Return a dictionary with the input micro link's PUIC as the key and the list of connected GraphMicroLinks as the value.
        return {input_micro_link.puic: output}

    def _get_micro_link_graph_mapping(self) -> Dict[str, List[GraphMicroLink]]:
        """Generates a mapping of micro links to their connected GraphMicroLinks in the graph.

        This method iterates over all micro links in the repository, obtaining connected micro links for each
        by processing both 'from_node' and 'to_node'. It combines these connections to form a comprehensive mapping
        that represents all possible micro link connections within the graph.

        Returns:
            Dict[str, List[GraphMicroLink]]: A dictionary mapping each micro link's PUIC to a list of connected GraphMicroLinks.
        """
        mapping_from = {}  # Initialize a dictionary to store connections originating from the 'from_node'.
        mapping_to = {}  # Initialize a dictionary to store connections originating from the 'to_node'.

        # Iterate over all micro links in the repository.
        for micro_link in self.micro_link_repo.micro_links:
            # Update the 'mapping_from' dictionary with connections found from the 'from_node' and 'from_node'
            mapping_from.update(self._get_linked_micro_links(micro_link, "from_node"))
            mapping_to.update(self._get_linked_micro_links(micro_link, "to_node"))

        # Combine the 'from' and 'to' mappings to create a comprehensive mapping.
        # This step merges connections from both directions for each micro link.
        common_dict = {key: mapping_from[key] + mapping_to[key] for key in mapping_from if key in mapping_to}

        # Return the comprehensive mapping of micro link connections.
        return common_dict

    def _get_micro_link_connections_repo(self, object_types: Optional[List[str]] = None) -> MicroLinkConnectionRepo:
        output: Dict[str, List[MicroLinkConnection]] = {}
        imx_objects = self.imx_situation.get_by_types(object_types) if object_types else self.imx_situation.get_all()

        for imx_object in imx_objects:
            if imx_object.rail_connection_infos:
                for rail_con_info in imx_object.rail_connection_infos.rail_infos:
                    # Use the new method to simplify creation of MicroLinkConnection objects.
                    connections = MicroLinkConnection.create_from_rail_info(rail_con_info, imx_object)
                    if rail_con_info.ref in output:
                        output[rail_con_info.ref].extend(connections)
                    else:
                        output[rail_con_info.ref] = connections

        # Sort the connections by measure for each reference key.
        for key, connections in output.items():
            output[key] = sorted(connections, key=lambda x: x.measure)

        return MicroLinkConnectionRepo(output)

    def build_graph(self) -> ImxGraph:
        """Builds the graph from the repositories."""
        [self.g.add_node(_.puic, pos=_.get_centroid_as_xy()) for _ in self.micro_link_repo.micro_links]
        mapping = self._get_micro_link_graph_mapping()

        for key, value in mapping.items():
            from_micro_link = self.micro_link_repo.get_by_puic(key)
            from_geometry_weight = from_micro_link.get_shapely().length / 2

            for item in value:
                to_geometry_weight = item.imx_micro_link.get_shapely().length / 2
                self.g.add_edge(
                    key,
                    item.imx_micro_link.puic,
                    weight=from_geometry_weight + to_geometry_weight,
                    current_direction=item.from_direction.name,
                    junction=item.junction.junctionRef,
                    next_direction=item.to_direction.name,
                    graph_junction=item.graph_junction,
                    jumper=item.jumper,
                )

        return ImxGraph(
            self.g, self.imx_situation, self.micro_link_repo, self.micro_node_repo, self.rail_connections_repo, self.micro_link_connections
        )



from typing import Iterable, List, Optional, Protocol

from lxml import etree as ET
from shapely.geometry.base import BaseGeometry

from imxInsights.domain.area.areaClassifiedTypeEnum import AreaClassifiedTypeEnum
from imxInsights.domain.area.imxProjectArea import ImxProjectArea
from imxInsights.domain.models.imxGeometry import ImxGeometry
from imxInsights.domain.models.imxMetadata import ImxObjectMetadata
from imxInsights.domain.models.ImxRailConnectionInfo import RailConnectionInfos
from imxInsights.domain.models.imxRelation import ImxReffedObjects
from imxInsights.repo.tree.imxTopoTreeObject import ImxTopoObject


class ImxObjectProtocol(Protocol):
    @property
    def properties(self) -> dict[str, str]:
        raise NotImplementedError

    @property
    def metadata(self) -> ImxObjectMetadata:
        raise NotImplementedError

    @property
    def geometry(self) -> Optional[ImxGeometry]:
        raise NotImplementedError

    @property
    def rail_connection_infos(self) -> Optional[RailConnectionInfos]:
        raise NotImplementedError

    @property
    def track_fragments(self) -> Optional[RailConnectionInfos]:
        raise NotImplementedError

    @property
    def demarcation_markers(self) -> Optional[RailConnectionInfos]:
        raise NotImplementedError

    @property
    def reffed_objects(self) -> ImxReffedObjects:
        raise NotImplementedError

    @property
    def topo_object(self) -> Optional[ImxTopoObject]:
        raise NotImplementedError

    @property
    def area(self) -> Optional[ImxProjectArea]:
        raise NotImplementedError

    @property
    def area_classify_type(self) -> Optional[AreaClassifiedTypeEnum]:
        raise NotImplementedError

    @property
    def parent(self) -> Optional["ImxObjectProtocol"]:
        raise NotImplementedError

    @property
    def children(self) -> List["ImxObjectProtocol"]:
        raise NotImplementedError

    @property
    def element(self) -> ET.Element:
        raise NotImplementedError

    @property
    def tag(self) -> str:
        raise NotImplementedError

    @property
    def path(self) -> str:
        raise NotImplementedError

    @property
    def puic(self) -> str:
        raise NotImplementedError

    @property
    def name(self) -> str:
        raise NotImplementedError

    @property
    def shapely(self) -> BaseGeometry:
        raise NotImplementedError

    # fmt: off
    def _parents_generator(self) -> Iterable["ImxObjectProtocol"]: ...  # noqa E704

    def register_topo_object(self, topo_object: ImxTopoObject): ...  # noqa E704

    def can_compare(self, other: Optional["ImxObjectProtocol"]) -> bool: ...  # noqa E704

    @staticmethod
    def lookup_tree_from_root_element(start: ET.Element) -> list["ImxObjectProtocol"]: ...  # noqa E704
    # fmt: on



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Iterable, List, Optional

from shapely import LineString, Point
from shapely.geometry.base import BaseGeometry

from imxInsights.diff.areaStatusEnum import AreaStatusEnum
from imxInsights.diff.change import Change, ChangeList
from imxInsights.diff.diffStatusEnum import DiffStatusEnum
from imxInsights.domain.area.imxProjectArea import ImxProjectArea
from imxInsights.repo.imxRepo import ImxObject, ObjectTree
from imxInsights.utils.log import logger
from imxInsights.utils.shapely_helpers import (
    gml_linestring_to_shapely,
    gml_point_to_shapely,
)


class ImxPropertyCompare:
    @classmethod
    def compare(cls, a: dict[str, str] | None, b: dict[str, str] | None) -> ChangeList:
        """
        Compares two flat dictionaries, detects changes & IMX specific changes (location).

        Args:
            a (Union[dict[str, str], None]): First dictionary to be compared.
            b (Union[dict[str, str], None]): Second dictionary to be compared.

        Returns:
            (ChangeList): A ChangeList object representing the changes between the two dictionaries.

        """
        keys = set(a.keys() if a is not None else []).union(b.keys() if b is not None else [])
        assert len(keys) > 0, "should find at lease 1 key in to compare"
        props = [cls._compare_property(key, a, b) for key in sorted(keys)]

        base_status = DiffStatusEnum.CREATED if a is None else DiffStatusEnum.DELETED if b is None else DiffStatusEnum.NO_CHANGE

        return ChangeList(status=base_status, properties=props)

    @classmethod
    def _compare_property(cls, key: str, a: dict[str, str] | None, b: dict[str, str] | None) -> Change:
        """
        Compares a single key in two dictionaries and returns a Change object representing the changes.

        Args:
            key (str): The key to be compared.
            a (Union[dict[str, str], None]): A dictionary to be compared.
            b (Union[dict[str, str], None]): A dictionary to be compared.

        Returns:
            (Change): A Change object representing the changes between the two values of the key.

        """
        val_a = a[key] if a is not None and key in a else None
        val_b = b[key] if b is not None and key in b else None

        def make_change(status: DiffStatusEnum) -> Change:
            return Change(key=key, status=status, a=val_a, b=val_b)

        if val_a is None:
            return make_change(DiffStatusEnum.CREATED)

        if b is None or key not in b or val_b is None:
            return make_change(DiffStatusEnum.DELETED)

        if val_a == val_b:
            return make_change(DiffStatusEnum.NO_CHANGE)

        if ".coordinates" in key:
            return make_change(cls._compare_gml_coordinates(key, val_a, val_b))

        return make_change(DiffStatusEnum.UPDATED)

    @classmethod
    def _compare_gml_coordinates(cls, key: str, coords_a: str, coords_b: str) -> DiffStatusEnum:
        """
        Compares two sets of coordinates and returns the status of the comparison.

        Args:
            key (str): The key of the coordinates to be compared.
            coords_a (str): The first set of coordinates.
            coords_b (str): The second set of coordinates.

        Returns:
            (DiffStatusEnum): A DiffStatus object representing the status of the comparison.

        """
        assert coords_a is not None and coords_b is not None
        if "Point.coordinates" in key:
            geom_a = gml_point_to_shapely(coords_a)
            geom_b = gml_point_to_shapely(coords_b)
        elif "LineString.coordinates" in key:
            geom_a = gml_linestring_to_shapely(coords_a)
            geom_b = gml_linestring_to_shapely(coords_b)
        else:
            logger.warning(f"Cannot determine if {key} is data upgrade, assuming changed if not exactly equal")
            return DiffStatusEnum.NO_CHANGE if coords_a == coords_b else DiffStatusEnum.UPDATED

        return cls._compare_geometry(geom_a, geom_b)

    @classmethod
    def _compare_geometry(cls, a: BaseGeometry, b: BaseGeometry, tolerance: Optional[float] = 4e-4) -> DiffStatusEnum:
        """
        Compares two geometries.

        Args:
            a (BaseGeometry): The first geometry object to compare.
            b (BaseGeometry): The second geometry object to compare.
            tolerance (Optional[float]): the tolerance level for determining if two geometries are equal, defaults to 4e-4.

        Returns:
            (DiffStatusEnum): The status of the comparison, whether the geometries are updated, not changed, or upgraded.

        """

        def get_values_to_compair(_a: BaseGeometry, _b: BaseGeometry) -> List[List[float], List[float], bool]:
            a_z, b_z = None, None
            z_status = 0
            match (_a.has_z, _b.has_z):
                case (False, False):
                    pass
                    # z_status = 0
                case (True, True):
                    # z_status = 0
                    if isinstance(_a, LineString) and isinstance(_b, LineString):
                        a_z = [item[2] for item in list(_a.coords)]
                        b_z = [item[2] for item in list(_b.coords)]
                    elif isinstance(_a, Point) and isinstance(_b, Point):
                        a_z = _a.z
                        b_z = _b.z
                    else:
                        raise ValueError("Should be same object types")

                case (True, False):
                    z_status = -1
                    if isinstance(_a, LineString):
                        a_z = [item[2] for item in list(_a.coords)]
                    elif isinstance(_a, Point):
                        a_z = _a.z
                    else:
                        raise ValueError("Should be same object types")

                case (False, True):
                    z_status = 1
                    if isinstance(_b, LineString):
                        b_z = [item[2] for item in list(_b.coords)]
                    elif isinstance(_b, Point):
                        b_z = _b.z
                        # b_z = _a.z
                    else:
                        raise ValueError("Should be same object types")

            return [a_z, b_z, z_status]

        def z_in_line_changed(_a_z: List[float], _b_z: List[float]) -> bool:
            for idx, item in enumerate(_a_z):
                if item != _b_z[idx]:
                    return True
            return False

        def z_value_changed(_a: BaseGeometry, _a_z: List[float], _b: BaseGeometry, _b_z: List[float]) -> bool:
            if isinstance(_a, LineString) or isinstance(_b, LineString):
                if len(_a_z) != len(_b_z):
                    return True
                else:
                    if z_in_line_changed(_a_z, _b_z):
                        return True
            if isinstance(_a, Point) or isinstance(_b, Point):
                if _a_z - _b_z != 0:
                    return True
            return False

        # TODO: make loads of local getting complex!

        a_z, b_z, z_status = get_values_to_compair(a, b)

        if a_z is not None and b_z is not None:
            if z_value_changed(a, a_z, b, b_z):
                return DiffStatusEnum.UPDATED

        if a.equals_exact(b, tolerance=tolerance):
            if z_status >= 0:
                return DiffStatusEnum.NO_CHANGE if z_status == 0 else DiffStatusEnum.UPGRADED

        return DiffStatusEnum.UPDATED


@dataclass(frozen=True)
class ImxObjectCompare:
    """
    Represents the comparison of two ImxObjects.

    Args:
        a (ImxObject): The first ImxObject to compare.
        b (ImxObject): The second ImxObject to compare.
        changes (ChangeList): The list of changes between the two objects.

    """

    a: ImxObject | None
    b: ImxObject | None
    changes: ChangeList = field(init=False)

    def __post_init__(self):
        assert self.a is not None or self.b is not None, "Must have at least one object"
        if self.a is not None:
            assert self.a.can_compare(self.b)

        a_props = self.a.properties if self.a is not None else None
        b_props = self.b.properties if self.b is not None else None
        changes = ImxPropertyCompare.compare(a_props, b_props)
        super().__setattr__("changes", changes)

    def __repr__(self) -> str:
        return (
            f"<ImxObjectCompare {self.path} name='{self.name}' puic={self.puic} " f"area_status={self.area_status} diff_status={self.diff_status}/>"
        )

    def __str__(self) -> str:
        parts = [self.path, self.name, str(self.area_status), self.puic]
        return "; ".join((part for part in parts if part is not None))

    @property
    def puic(self) -> str:
        """Returns the puic of the ObjectComparison."""
        return self._any().puic

    @property
    def tag(self) -> str:
        """Returns the tag of the ObjectComparison."""
        return self._any().tag

    @property
    def path(self) -> str:
        """Returns the path of the ObjectComparison."""
        return self._any().path

    @property
    def diff_status(self) -> DiffStatusEnum:
        """Returns the diff status of the ObjectComparison."""
        return self.changes.status

    @property
    def name(self) -> str | None:
        """Returns the Combined name, '<NewName> (<OldName>)' or just the '<Name>' if same."""
        if self.only_one:
            name = self._any().name
            return name if name is not None else None

        assert self.a is not None and self.b is not None
        a = self.a.name
        b = self.b.name
        return f"*{b} {f'({a})' if len(a) > 0 else ''}".rstrip() if a != b else a

    @property
    def area_a(self) -> ImxProjectArea | None:
        """Returns the area of `a`."""
        return self.a.area if self.a is not None else None

    @property
    def area_b(self) -> ImxProjectArea | None:
        """Returns the area of `b`."""
        return self.b.area if self.b is not None else None

    @property
    def area_status(self) -> AreaStatusEnum:
        """Determines the status of the comparison of two ImxObjects based on their areas."""
        if self.area_a is None and self.a is not None or self.area_b is None and self.b is not None:
            return AreaStatusEnum.INDETERMINATE

        elif self.area_a is None and self.area_b is not None:
            return AreaStatusEnum.CREATED

        elif self.area_a is not None and self.area_b is None:
            return AreaStatusEnum.DELETED

        elif self.area_a.name == self.area_b.name:
            return AreaStatusEnum.NO_CHANGE

        elif self.area_a.name != self.area_b.name:
            return AreaStatusEnum.MOVED

        raise ValueError(f"Unknown area combination: A={self.area_a}, B={self.area_b}")

    @property
    def only_one(self) -> bool:
        """Returns a boolean value indicating whether only one object is present."""
        return (self.a is None) ^ (self.b is None)

    def _any(self) -> ImxObject:
        if self.a is not None:
            return self.a

        assert self.b is not None, "Should have a or b"
        return self.b

    @staticmethod
    def object_tree_factory(a: ObjectTree, b: ObjectTree) -> Iterable[ImxObjectCompare]:
        """
        Compares two object trees and returns an iterable of ObjectComparison instances representing the comparison between the two trees.

        Args:
            a (ObjectTree): The second object tree to compare.
            b (ObjectTree): The second object tree to compare.

        Returns:
            (Iterable[ImxObjectCompare]): An iterable of ObjectComparison instances representing the comparison between the two trees.

        """
        all_keys = a.keys.union(b.keys)
        for key in all_keys:
            match_a = a.find(key)
            match_b = b.find(key)
            yield ImxObjectCompare(match_a, match_b)

    @staticmethod
    def imx_object_list_factory(a: List[ImxObject], b: List[ImxObject]) -> Iterable[ImxObjectCompare]:
        """
        Generates `ObjectComparison` instances by comparing two lists of `ImxObject` instances.

        If an `ImxObject` instance is present in both lists, it is compared, otherwise it is compared to `None`.

        Args:
            a (List[ImxObject]): The first list of `ImxObject` instances to compare.
            b (List[ImxObject]): The second list of `ImxObject` instances to compare.

        Returns:
            (Iterable[ImxObjectCompare]): An iterable of `ObjectComparison` instances.

        """
        dict_a = {f"{item.puic}": item for item in a}
        dict_b = {f"{item.puic}": item for item in b}
        for puic, value in dict_a.items():
            if puic in dict_b.keys():
                yield ImxObjectCompare(value, dict_b[puic])
            else:
                yield ImxObjectCompare(value, None)

        for puic, value in dict_b.items():
            if puic not in dict_a.keys():
                yield ImxObjectCompare(value, None)



from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, List, Optional, TypeVar

from lxml.etree import Element
from shapely import LineString, Point

from imxInsights.domain.models.imxRelation import ImxRelation
from imxInsights.graph.imxGraphJunction import ImxGraphJunction, ImxGraphSwitchPosition
from imxInsights.repo.tree.objectTreeProtocol import ImxObjectProtocol
from imxInsights.utils.log import logger
from imxInsights.utils.shapely_geojson import GeoJsonFeature, GeoJsonFeatureCollection
from imxInsights.utils.shapely_helpers import ShapelyTransform

T = TypeVar("T")


def parse_to_bool_if_unknown_true(string):
    string_lower = string.lower()
    if string_lower == "true":
        return True
    elif string_lower == "false":
        return False
    elif string_lower == "unknown":
        return True
    else:
        raise ValueError("Invalid boolean string: {}".format(string))


@dataclass
class ImxRailConnection:
    imx_object: ImxObjectProtocol
    puic: str = field(init=False)
    passage_refs: List[ImxRelation] = field(init=False)

    def __post_init__(self):
        self.puic = self.imx_object.puic
        try:
            self.passage_refs = [_ for _ in self.imx_object.reffed_objects.objects if _.type in ["passageRefs", "PassageRefs"]]
        except Exception as e:
            print(e)


@dataclass
class ImxMicroNodeJumper:
    _element: Element
    from_idx: str = field(init=False)
    to_idx: str = field(init=False)
    is_traversible: bool = field(init=False)
    is_two_way: bool = field(init=False)
    passage_refs: List[str] = field(default_factory=list)

    def __post_init__(self):
        self.from_idx = self._element.attrib.get("fromIndex")
        self.to_idx = self._element.attrib.get("toIndex")
        self.is_traversible = parse_to_bool_if_unknown_true(self._element.attrib.get("isTraversible"))
        self.is_two_way = parse_to_bool_if_unknown_true(self._element.attrib.get("isTwoWay"))

        # todo: make explicit check what to use
        try:
            self.passage_refs = self._element.attrib.get("passageRefs").split(" ")
        except Exception as e:
            try:
                self.passage_refs = self._element.find(".//{*}PassageRefs").text.split(" ")
            except Exception as e_2:
                print(e, e_2)


@dataclass
class ImxMicroNode:
    _imx_obj: ImxObjectProtocol
    puic: str = field(init=False)
    junctionRef: str = field(init=False)
    jumpers: List[ImxMicroNodeJumper] = field(default_factory=list)

    def __post_init__(self):
        self.puic = self._imx_obj.puic
        self.junctionRef = self._imx_obj.puic.split("_")[0]
        self.jumpers = [ImxMicroNodeJumper(_) for _ in self._imx_obj.element.findall(".//{http://www.prorail.nl/IMSpoor}Jumper")]


@dataclass
class ImxMicroLinkFromOrToNode:
    _element: Element
    ref: str = field(init=False)
    idx: int = field(init=False)
    micro_node: ImxMicroNode | None = None
    tag: str = field(init=False)

    def __post_init__(self):
        self.ref = self._element.attrib.get("nodeRef")
        self.idx = self._element.attrib.get("portIndex")
        self.tag = self._element.tag.split("}")[1]


@dataclass
class ImxMicroLink:
    _imx_obj: ImxObjectProtocol
    from_node: ImxMicroLinkFromOrToNode = field(init=False)
    to_node: ImxMicroLinkFromOrToNode = field(init=False)
    _rail_connection: ImxObjectProtocol = field(init=False)
    puic: str = field(init=False)
    track_ref: str = field(init=False)
    passage_refs: List[str] = field(init=False)

    def __post_init__(self):
        self.puic = self._imx_obj.puic.split("_")[0]
        self.from_node = ImxMicroLinkFromOrToNode(self._imx_obj.element.find(".//{http://www.prorail.nl/IMSpoor}FromMicroNode"))
        self.to_node = ImxMicroLinkFromOrToNode(self._imx_obj.element.find(".//{http://www.prorail.nl/IMSpoor}ToMicroNode"))
        self._rail_connection = self._imx_obj.reffed_objects.objects[0].reffed_object
        if self._rail_connection is not None:
            self.track_ref = self._rail_connection.properties["@trackRef"] if "@trackRef" in self._rail_connection.properties.keys() else ""
            self.passage_refs = [_.key for _ in self._rail_connection.reffed_objects.objects if _.type in ["passageRefs", "PassageRefs"]]
        else:
            logger.critical(f"MicroLink {self.puic} object does not have a railConnection counterpart, the app will crash: fix issue and try again!")
            self.track_ref = ""
            self.passage_refs = []

    def link_micro_nodes(self, micro_nodes: dict[str, ImxMicroNode]):
        self.from_node.micro_node = micro_nodes[self.from_node.ref]
        self.to_node.micro_node = micro_nodes[self.to_node.ref]

    def get_centroid_as_xy(self) -> tuple[float, float]:
        x = self._rail_connection.shapely.centroid.x
        y = self._rail_connection.shapely.centroid.y
        return x, y

    def get_shapely(self):
        return self._rail_connection.shapely


class DirectionEnum(Enum):
    UPSTREAM = "upstream"
    DOWNSTREAM = "downstream"
    UNKNOWN = "unknown"


@dataclass
class GraphMicroLink:
    imx_micro_link: ImxMicroLink
    from_direction: DirectionEnum
    to_direction: DirectionEnum
    junction: ImxMicroNode
    graph_junction: ImxGraphJunction
    jumper: ImxMicroNodeJumper


class Repo:
    @staticmethod
    def _check_and_return(list_of_things: List[T]) -> T:
        if len(list_of_things) == 1:
            return list_of_things[0]
        if len(list_of_things) > 1:
            raise ValueError("Multiple instances found in repo")
        raise ValueError("Not Found")


@dataclass
class RailConnectionRepo(Repo):
    rail_connections: List[ImxRailConnection] = field(default_factory=list)

    def get_by_puic(self, puic: str) -> ImxRailConnection:
        _ = [item for item in self.rail_connections if item.puic == puic]
        return self._check_and_return(_)


@dataclass
class MicroNodeRepo(Repo):
    micro_nodes: List[ImxMicroNode] = field(default_factory=list)

    def get_by_puic(self, puic: str) -> ImxMicroNode:
        _ = [item for item in self.micro_nodes if item.junctionRef == puic]
        return self._check_and_return(_)


@dataclass
class MicroLinkRepo(Repo):
    micro_links: List[ImxMicroLink] = field(default_factory=list)

    def get_by_puic(self, puic: str) -> ImxMicroLink:
        _ = [item for item in self.micro_links if item.puic == puic]
        return self._check_and_return(_)


class MicroLinkConnectionTypeEnum(Enum):
    POINT = "point"
    LINE_START = "line_start"
    LINE_END = "line_end"


@dataclass
class MicroLinkConnection:
    measure: float
    ref: str
    direction: str
    connection_type: MicroLinkConnectionTypeEnum
    item: any  # Assuming 'item' can be of any type for now

    @staticmethod
    def create_from_rail_info(rail_con_info, imx_object):
        """Creates MicroLinkConnection instances from rail connection information.

        Args:
            rail_con_info: The rail connection information object.
            imx_object: The IMX object associated with the rail connection.

        Returns:
            List[MicroLinkConnection]: A list containing one or more MicroLinkConnection instances.
        """
        connections = []
        if hasattr(rail_con_info, "at_measure"):
            # Handle point-type rail connections.
            connections.append(
                MicroLinkConnection(
                    rail_con_info.at_measure, rail_con_info.ref, rail_con_info.direction, MicroLinkConnectionTypeEnum.POINT, imx_object
                )
            )
        else:
            # Handle line-type rail connections, assuming they have 'from_measure' and 'to_measure'.
            # todo: check if present, else give error warning.
            if not hasattr(rail_con_info, "from_measure") or not hasattr(rail_con_info, "to_measure"):
                pass
            else:
                connections.extend(
                    [
                        MicroLinkConnection(
                            rail_con_info.from_measure,
                            rail_con_info.ref,
                            rail_con_info.direction,
                            MicroLinkConnectionTypeEnum.LINE_START,
                            imx_object,
                        ),
                        MicroLinkConnection(
                            rail_con_info.to_measure,
                            rail_con_info.ref,
                            rail_con_info.direction,
                            MicroLinkConnectionTypeEnum.LINE_END,
                            imx_object,
                        ),
                    ]
                )
        return connections


@dataclass
class MicroLinkConnectionRepo:
    micro_links: Dict[str, list[MicroLinkConnection]] = field(default_factory=list)

    def get_by_puic(self, puic: str) -> List[MicroLinkConnection]:
        return self.micro_links[puic]

    def get_keys(self):
        return [_ for _ in self.micro_links.keys()]


@dataclass
class PathEdgeData:
    source_direction: DirectionEnum
    target_direction: DirectionEnum
    junction_ref: str
    graph_junction: ImxGraphJunction
    jumper: ImxMicroNodeJumper

    def as_string(self):
        return f"source: {self.source_direction}, target:{self.target_direction} true_node:{self.junction_ref}"


@dataclass
class PathEdge:
    source: str
    target: str
    data: Optional[PathEdgeData] = None

    @property
    def edge_geometry(self) -> LineString:
        return self.data.graph_junction.get_path(self.data.jumper.passage_refs).geometry

    @property
    def switches_on_edge(self) -> list[ImxGraphSwitchPosition]:
        return self.data.graph_junction.get_path(self.data.jumper.passage_refs).switches_in_path


@dataclass
class PathEdgeSwitchMechanismDisplay:
    ref: str
    name: str
    geometry: Point
    position: str
    path_geometry: LineString


@dataclass
class GraphRoute:
    path: List[PathEdge] = field(default_factory=list)
    geometry: LineString = None
    objects_on_route: List[Any] = field(default_factory=list)
    # todo: get switches in route by pathEdges
    #  switches_on_route: List[Any] = field(default_factory=list)

    @property
    def switches_on_route(self) -> list[PathEdgeSwitchMechanismDisplay]:
        output = []
        for path_edge in self.path:
            for switch_on_edge in path_edge.switches_on_edge:
                output.append(
                    PathEdgeSwitchMechanismDisplay(
                        switch_on_edge.switch_mechanism.puic,
                        switch_on_edge.switch_mechanism.name,
                        switch_on_edge.switch_mechanism.geometry_reprojected,
                        switch_on_edge.position.value,
                        path_edge.edge_geometry,
                    )
                )

        return output

    @property
    def lr_string(self) -> str:
        return "-".join([f"{_.name}_{_.position[0].upper()}" for _ in self.switches_on_route])

    def as_shapely(self) -> LineString:
        return self.geometry

    def as_feature_collection(self):
        route_feature = GeoJsonFeature([ShapelyTransform.rd_to_wgs(self.geometry)], {"path": self.lr_string, "stroke-width": 20.0, "color": "yellow"})

        switch_jumper_features = [
            GeoJsonFeature(
                [ShapelyTransform.rd_to_wgs(item.path_geometry)],
                {"switch_name": item.name, "position": item.position, "stroke-width": 25.0, "color": "orange"},
            )
            for item in self.switches_on_route
        ]

        switch_mech_features = [
            GeoJsonFeature(
                [ShapelyTransform.rd_to_wgs(item.geometry)], {"switch_name": item.name, "position": item.position, "imx_type": "SwitchMechanism"}
            )
            for item in self.switches_on_route
        ]

        # 	"featureTags": ["tag"]
        # 	an array of string tags that are used for hierarchical filtering;
        # 	featureTags values can have a pipe (|) inside thus making groups on different levels;
        # 	for example, "group_1|group_2|group_3" will result in a three-level deep tree.

        # # todo: make sure it's only in direction of travel or none
        features = [
            GeoJsonFeature([ShapelyTransform.rd_to_wgs(item.item.shapely)], item.item.properties | {"imx_type": item.item.path})
            for item in self.objects_on_route
        ]

        # features = []
        features.extend(switch_jumper_features)
        features.extend(switch_mech_features)
        features.append(route_feature)
        return GeoJsonFeatureCollection(features)



from dataclasses import dataclass, field
from pathlib import Path

from lxml import etree as ET

NS_IMSPOOR = "http://www.prorail.nl/IMSpoor"
NS_GML = "http://www.opengis.net/gml"


@dataclass(frozen=True)
class XmlFile:
    """
    Represents an XML file, providing methods for parsing and checking its existence.

    This class allows you to work with XML files by providing a convenient interface for
    parsing and checking the existence of an XML file. It can be used to load and parse
    an XML file, and to determine whether the file exists.

    Args:
        path (Path): The path to the XML file.
        root (ET.ElementTree, optional): An optional pre-parsed XML root element. Default is None.

    Attributes:
        path (Path): The path to the XML file.
        root (ET.ElementTree): The parsed XML root element. It is set to None initially and
            will be parsed when necessary using the `ET.parse` method.

    Raises:
        ValueError: If the provided `path` does not exist or is not a file.

    Note:
        When parsing XML files, this class uses an XMLParser with comments removed to ensure
        that comments are not parsed as nodes. However, please note that comments will be lost
        when saving the XML file.

    Examples:
        >>> xml_path = Path("example.xml")
        >>> xml_file = XmlFile(xml_path)
        >>> xml_file.exists
        True
        >>> xml_file.root  # The XML root element is automatically parsed when accessed.
        <Element 'root' at 0x7f5de6f18d10>

    """

    path: Path
    root: ET.ElementTree = field(kw_only=True, hash=False, repr=False, default=None)
    # tag: str | None = None

    def __post_init__(self) -> None:
        if self.root is not None:
            return

        if not self.exists:
            raise ValueError(f"Invalid path {self.path}")

        # mutcho use of comments for local purposes ET will parse them as nodes so make sure comments are not parsed...
        # when saving xml comments are lost!
        parser = ET.XMLParser(remove_comments=True)
        super().__setattr__("root", ET.parse(self.path, parser))

    @property
    def exists(self) -> bool:
        return self.path.exists() and self.path.is_file()


def trim_tag(tag: ET._Element | str) -> str:
    if isinstance(tag, ET._Element):
        tag = str(tag.tag)
    return tag.split("}")[1] if "}" in tag else tag


def find_base_entity(elem: ET._Element) -> ET._Element:
    while elem is not None and "puic" not in elem.keys():
        elem = elem.getparent()
    if elem is None:
        raise ValueError
    return elem


def find_parent_entity(elem: ET._Element) -> ET._Element | None:
    assert "puic" in elem.keys(), "Element has no puic!"
    try:
        parent = find_base_entity(elem.getparent())
    except ValueError:
        return None

    return parent if trim_tag(parent.tag) != "Project" else None


def find_coordinates(elem: ET._Element) -> list[ET._Element]:
    target = f"{{{NS_GML}}}coordinates"
    if elem.tag == target:
        return [elem]

    matches: list[ET._Element] = elem.findall(f".//{target}")

    # filter out coordinates of child entities
    # assert "puic" in elem.keys(), "Cannot find coordinates of non-entity"
    return [coord for coord in matches if find_base_entity(coord) == elem]


def find_puic(elem: ET._Element) -> str:
    return str(find_base_entity(elem).get("puic"))


def lxml_element_to_dict(node: ET.Element, attributes: object = True, children: object = True) -> dict[str, dict | str | list]:
    """
    Convert lxml.etree node into a dict. adapted from https://gist.github.com/jacobian/795571.

    Args:
        node (ET.Element): the lxml.etree node to convert into a dict
        attributes (Optional[bool]): include the attributes of the node in the resulting dict, defaults to True

    Returns:
        ( dict[str, dict | str | list]): A dictionary representation of the lxml.etree node

    """
    result = dict[str, dict | str | list]()
    if attributes:
        for key, value in node.attrib.items():
            result[f"@{key}"] = value

    if not children:
        return result

    for element in node.iterchildren():
        key = trim_tag(element)

        # Process element as tree element if the inner XML contains non-whitespace content
        if element.text and element.text.strip():
            value = element.text
        else:
            value = lxml_element_to_dict(element)

        if key in result:
            match = result[key]
            if isinstance(match, list):
                match.append(value)
            else:
                result[key] = [match, value]
        else:
            result[key] = value

    return result
